function [Bs, alphas, Ss, Vs, Fs, mus, rankBs, clstnoBs, objvals] = ...    
    Func_CRL_SolPath(X, Y, qGridInfo, rGridInfo, numInits, lossType, estimateIntercept,  ...
                     multi_start_params, convCtrl, linesearchCtrl)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This function generates a solution path for CRL
%
% Input arguments:
%   X: the design matrix/matrices (no intercept column for ordinary matrix models, but may contain it for sample-based trace models)
%   Y: the response matrix (Y ~ XB + 1 * alpha'), 
%      or the response vector for trace models, either the general one 
%      y_i (scalar) ~ < X(:, :, i), B > + alpha_i or a "sample-based trace model": 
%      y_i (scalar) ~ < X(i, :), B(i, :) >  + alpha_i
%   qGridInfo: A structure array to obtain a cluster-number grid. 
%              Note: If the field 'qGrid' is provided, the field 'qPoints' is 
%                 ignored. If both are empty, we will construct a q-grid in
%                 the default manner
%       -> qGrid: provided by the user. If empty, use the default: an evenly spaced "qPoints many" cluster 
%                 numbers ranging from 1 to min(n, p)).
%       -> qPoints: provided by the user. If empty, use the default (20). 
%   rGridInfo: A structure array to obtain a rank grid.
%              Note: If the field 'rGrid' is provided, the field 'rPoints' is 
%                   ignored. If both are empty, we will construct a r-grid by default.
%       -> rGrid: provided by the user. If empty, use the default: an evenly spaced "rPoints many" ranks 
%                 ranging from 1 to min(n, p, m)).
%       -> rPoints: provided by the user. If empty, use the default (8).
%   numInits: Check Func_multipleinits for details.
%   losstype: Could be 'gaussian', 'bernoulli', 'poisson', a robust loss 
%             (or any other loss of interest, if implemented by the user). 
%             Check Func_Loss_Grad for more details.
%   estimateIntercept: (Default: 1)
%       0: No intercept estimation (return zeros).
%       1: The intercept alpha will be simultaneously estimated.  
%   multi_start_params: Check Func_multipleinits for details.
%   convCtrl: Check Func_multipleinits for details.
%   linesearchCtrl: Check Func_multipleinits for details.
%
% Output arguments:
%   Bs: A 3D array of coefficient matrix estimate solution path.
%   alphas: A matrix of intercept estimate solution path.
%   Ss: A cell array of loading matrix solution path.
%   Vs: A cell array of rotation matrix solution path.
%   Fs: A cell array of binary membership matrix solution path.
%   mus: A cell array of cluster centroid matrix solution path.
%   rankBs: A vector ranks solution path.
%   clstnoBs: A vector of cluster numbers solution path.
%   objvals: A vector of objective function value solution path.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if ~contains(lossType, 'tracemodel')
    [n, m] = size(Y); p = size(X, 2); 
    intrcptNo = m; % num of components in the intercept column
    temp_rUppBnd = min([rank(X, 1e-5), m]); % rankX = rank(X, 1e-5);
else
    if size(X, 3) > 1
        % The most general trace model: y_i (scalar) ~ < X(:, :, i), B > + alpha_i
        [n] = size(Y, 1); 
        p = size(X, 1); m = size(X, 2); % B is of size pxm
        intrcptNo = n; % num of components in the intercept column
    else
        % if size(X, 3) == 1, i.e., a two-dim design matrix, 
        % we assume a "sample-based trace model": y_i (scalar) ~ < X(i, :), B(i, :) >  + alpha_i
        [n] = size(Y, 1); p = size(X, 1); m = size(X, 2); % B is of size pxm (or nxp in this special case)
        intrcptNo = n;
    end
    temp_rUppBnd = min([n, m, p]); 
end

debug_output = 0; % 1; % For displaying some useful info.

if isempty(numInits)
    % default CRL initial values parameters
    numInits = struct('rrr', [], 'rand', [50]); % struct('rrr', [], 'rand', []); 
end
if isempty(multi_start_params)
    % default Multi-start parameters in CRL
    multi_start_params = struct('NumsStarts', [], 'ItersForPrejudgement', []);
    multi_start_params.NumsStarts = [-1, 3, 1]; %[-1, 3, 1]; % number of initial vals used in prejudgement stages. //No matter what the 1st element is, in the first stage, all initial values will be used. Also, note that the last element is the number of starting values for the final optimization stage
    multi_start_params.ItersForPrejudgement = [15, 15]; %[3, 3]; %  ith entry: the number of iterations for the ith (prejudgement) stage. // If not enough elements are provided, we will repeat the last provided value to fill in the whole array
end
if isempty(convCtrl)
    % default CRL convergence control parameters
    convCtrl = struct('outer', [], 'inner', []);
    convCtrl.outer = struct('maxiter', [], 'B_tol', []);
    convCtrl.inner = struct('maxiter', [], 'B_tol', [], 'func_tol', [], 'F_tol', []);
    %------outer loop stopping criterion------------%
    convCtrl.outer.maxiter = 500; % 20, 50 % Maximum outer iteration allowed
    convCtrl.outer.B_tol = 1e-05; % error tolerance in change of B (outer loop)
    %------inner loop stopping criterion------------%
    convCtrl.inner.maxiter = 10; % Maximum inner iteration allowed
    convCtrl.inner.B_tol = 1e-03; % error tolerance in change of B (inner loop)
    convCtrl.inner.func_tol = 1e-03; % error tolerance in change of function value
    convCtrl.inner.F_tol = 0; % Using stability in membership matrix as convergence criterion
end
if isempty(linesearchCtrl)
    % default CRL line search criterion
    linesearchCtrl.rho_redfactor = 0.5; % 0.25 % 0.75
    linesearchCtrl.numRhosInTrial = 5; % 1: no search
    linesearchCtrl.linesearchDir = 'adaptive';%'backward'; %  
end

% Construction manners of q-grid and r-grid:
deflt_qPoints = 20; % default size of q-grid
deflt_rPoints = 8;  % default size of r-grid
if ~exist('qGridInfo', 'var') || isempty(qGridInfo)
    qGrid_manner = 'default'; % We will construct a default one later
else
    if isfield(qGridInfo, 'qGrid') && ~isempty(qGridInfo.qGrid)
        qGrid_manner = 'user_grid';   % higher priority than 'user_points'
    elseif isfield(qGridInfo, 'qPoints') && ~isempty(qGridInfo.qPoints)
        qGrid_manner = 'user_points';
    else
        qGrid_manner = 'default'; % We will construct a default one later
    end
end
if ~exist('rGridInfo', 'var') || isempty(rGridInfo)
    rGrid_manner = 'default'; % We will construct a default one later
else
    if isfield(rGridInfo, 'rGrid') && ~isempty(rGridInfo.rGrid)
        rGrid_manner = 'user_grid';   % higher priority than 'user_points'
    elseif isfield(rGridInfo, 'rPoints') && ~isempty(rGridInfo.rPoints)
        rGrid_manner = 'user_points';
    else
        rGrid_manner = 'default'; % We will construct a default one later
    end
end

% Now construct the q-grid (q: the number of distinct rows in B which is of size pxm)
temp_qUppBnd = min([n, p]); temp_qLowBnd = 1;
if strcmp(qGrid_manner, 'user_grid')
    qGrid = unique( sort(qGridInfo.qGrid, 'ascend') );
    qGrid(qGrid > temp_qUppBnd) = []; qGrid(qGrid < temp_qLowBnd) = [];
    qGridlen = numel(qGrid); % the max # of q values in the q-grid
else
    if strcmp(qGrid_manner, 'user_points') && ~isempty(qGridInfo.qPoints) 
        qPoints = qGridInfo.qPoints; % specified by the user
    else
        qPoints = deflt_qPoints;
    end    
    qGrid = unique( round(linspace(temp_qLowBnd, temp_qUppBnd, qPoints)) );
    qGridlen = numel(qGrid); % the max # of q values in the q-grid
end    
if numel(qGrid) == 0
    error('No effective value for cluster number provided');
end

% If necessary, construct a raw r-grid (NOT varying with q)
if strcmp(rGrid_manner, 'user_grid')
    temp_rLowBnd = 1;
    rGrid = unique( sort(rGridInfo.rGrid, 'ascend') );
    rGrid(rGrid > temp_rUppBnd) = temp_rUppBnd; % []; % If the specified upper bound is too large, truncate it to the nontrivial r bound
    rGrid(rGrid < temp_rLowBnd) = []; % note that r is the upper bound parameter in the constraint
    rGridlen = numel(rGrid);  
    if numel(rGrid) == 0
        warning('No effective values of rank provided by the user! Use the default rank grid.');
        if ~isempty(rGridInfo.rPoints) 
            rPoints = rGridInfo.rPoints; % specified by the user
        else
            rPoints = deflt_rPoints;
        end
        rGridlen = rPoints; % the max # of r values in the r-grid
    end
else
    if strcmp(rGrid_manner, 'user_points') && ~isempty(rGridInfo.rPoints) 
        rPoints = rGridInfo.rPoints;
    else
        rPoints = deflt_rPoints;
    end
    rGridlen = rPoints; % the max # of r values in the r-grid
    rGrid = []; % We will construct an *adaptive* rGrid according to the value of q of interest
end


% Whether or not we will remove redundant and trivial solutions along the path
complexRem = true;
redundRem = true;
zeroRem =true;


%----------Start the solution path calculation-------------%
Bs = zeros(p, m, qGridlen*rGridlen); % use the largest possible sizes
alphas = zeros(intrcptNo, size(Bs, 3));
Ss = cell(1, size(Bs,3));
Vs = cell(1, size(Bs,3));
Fs = cell(1, size(Bs,3));
mus = cell(1, size(Bs,3));
rankBs = - ones(1, size(Bs, 3));
clstnoBs = - ones(1, size(Bs, 3));
numBs = 0; 
for q = qGrid
    if ~isempty(rGrid)
        % When r > q, peform a necessary truncation 
        rGrid_given_q = rGrid(rGrid <= q); % might become empty 
    end    
    if isempty(rGrid)
        temp_rUppBnd2 = min([q, temp_rUppBnd]); temp_rLowBnd = 1; 
        rGrid_given_q = unique( round(linspace(temp_rLowBnd, temp_rUppBnd2, rPoints)) ); % the grid length might be < rPoints
    end
    if isempty(rGrid_given_q), rGrid_given_q = [q]; end
     
    for r = rGrid_given_q  
        numBs = numBs + 1;
        if debug_output
            disp(['Calculating the CRL solution at q = ', num2str(q), ', r = ', num2str(r)])
        end
        
        % Get the solution associated with the specified (q, r)
        [B_est, alpha_est, S_est, V_est, F_est, mu_est]  = ...
            Func_CRL_multipleinits(X, Y, q, r, [], numInits, lossType, estimateIntercept,...
                                 multi_start_params, convCtrl, linesearchCtrl);
                             
        rankBs(numBs) = rank(B_est, 1e-5);
        clstnoBs(numBs) = size( unique(S_est, 'rows'), 1);
        
        Bs(:, :, numBs) = B_est;
        alphas(:, numBs) = alpha_est;
        Fs{numBs} = F_est;
        Ss{numBs} = S_est;
        Vs{numBs} = V_est;
        mus{numBs} = mu_est;
        
        % Make a judgement to terminate the loop earlier, if the current
        % model is already too complex
        if r < max(rGrid_given_q) && ( ( (clstnoBs(numBs) + m) * rankBs(numBs) > n * m * .8 && ~contains(lossType, 'tracemodel') ) || ...
                ( (clstnoBs(numBs) + m) * rankBs(numBs) > n * .8 && contains(lossType, 'tracemodel') ) ) % .5??? %
            disp('Model already seems to be complex enough. Stop increasing r.');
            break;
        end
        
    end

end



if complexRem 
    % Make sure the estimates have the # of groups less than temp_qUppBnd %%the rank of X in matrix models
    validModelInds = find(clstnoBs <= temp_qUppBnd); % rankX);
    numBs = numel(validModelInds);
    Bs = Bs(:, :, validModelInds); alphas = alphas(:, validModelInds);
    Ss = Ss(validModelInds); Vs = Vs(validModelInds); 
    Fs = Fs(validModelInds); mus = mus(validModelInds);
    rankBs = rankBs(validModelInds); clstnoBs = clstnoBs(validModelInds);    
end
if redundRem 
    % Remove redundant Bs
    redundantInds = find(sum(sum(abs(Bs(:,:,2:end)-Bs(:,:,1:end-1)))) <1e-4) + 1;
    Bs(:, :, redundantInds) = []; alphas(:, redundantInds) = [];
    Ss(redundantInds) = []; Vs(redundantInds) = []; 
    Fs(redundantInds) = []; mus(redundantInds) = [];
    rankBs(redundantInds) = []; clstnoBs(redundantInds) = [];
    numBs = numel(rankBs);
end
if zeroRem 
    % Remove the zero rank est
    zerorankInds = find(rankBs == 0);
    Bs(:, :, zerorankInds) = []; alphas(:, zerorankInds) = [];
    Ss(zerorankInds) = []; Vs(zerorankInds) = []; 
    Fs(zerorankInds) = []; mus(zerorankInds) = [];
    rankBs(zerorankInds) = []; clstnoBs(zerorankInds) = [];
    numBs = numel(rankBs);
end

if rankBs(end) == -1
    % In this case, the pre-specified grid was large and the redundRem checking only removes next == current. So we need to remove one additional empty slot. 
    numBs = numBs - 1;
end

% The final solution path after getting rid of redundant estimates
Bs = Bs(:, :, 1:numBs); alphas = alphas(:, 1:numBs);
Ss = Ss(1:numBs); Vs = Vs(1:numBs); 
Fs = Fs(1:numBs); mus = mus(1:numBs);
rankBs = rankBs(1:numBs); clstnoBs = clstnoBs(1:numBs);

% Calculate the associated objective function value
objvals = -ones(1, numBs);
for indB = 1:numBs
    B_est = Bs(:, :, indB); alpha_est = alphas(:, indB);
%     [Theta] = Func_Eval_Theta_SysComp(X, B_est, alpha_est, lossType); % For ordinary models, Theta = X * B_est + alpha_est';
%     objvals(indB) = Func_Loss_Grad(Y, Theta, lossType, 0); %0: loss value only
    objvals(indB) = Func_Loss_Grad_Balpha(Y, X, B_est, alpha_est, lossType, 0); %0: loss value only
end

end
