function [B_est, alpha_est, S_est, V_est, F_est, mu_est]  = ...
    Func_CRL_multipleinits(X, Y, q, r, init, numInits, lossType, estimateIntercept,...
                              multi_start_params, convCtrl, linesearchCtrl)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This function solves supervised CRL problem for given multiple 
% initialization points. If multiple initialization is not given it constructs
% initializations and runs the algorithm. Check multi_start_params argument
% for more details.
%
% Input arguments:
%   X: the design matrix/matrices (no intercept column for ordinary matrix models, but may contain it for sample-based trace models)
%   Y: the response matrix (Y ~ XB + 1 * alpha'), 
%      or the response vector for trace models, either the general one 
%      y_i (scalar) ~ < X(:, :, i), B > + alpha_i or a "sample-based trace model": 
%      y_i (scalar) ~ < X(i, :), B(i, :) >  + alpha_i
%   q: the cluster size of interest.
%   r: the rank value of interest.
%   init: Structure array containing initial values of F, mu, V. (and
%         of alpha if estimateIntercept = 1). Once given, "numInits"
%         becomes inactive.
%       -> init. V: Initial rotation matrix.
%       -> init. F: Initial binary membership matrix.
%       -> init. mu: Initial cluster centroid matrix.       
%       -> init. alpha: Initial intercept column vector.
%   numInits: the user can specify this number struct instead of providing
%           the specific values in "init". But once init is given,
%           "numInits" becomes inactive. (Default: 20+2 initial points in total)
%       -> numInits.rrr for RRR-type initial values: Array of desired ranks to be used for intials 
%                         that are obtained from RRR. If empty, use the default [q, m]. If -1, no RRR initial values
%       -> numInits.rand for rand-type initial values: Number of random initial values (based on subset sampling). 
%                           Default: 20. Can be zero which means no random initial values will be used.
%   losstype: Could be 'gaussian', 'bernoulli', 'poisson', a robust loss, 
%             or any other loss of interest (not all implemented yet). 
%             Check Func_Loss_Grad for more details.
%   estimateIntercept: (Default: 1)
%       0: No intercept estimation (return zeros).
%       1: The intercept alpha will be simultaneously estimated.  
%   multi_start_params: A structure array to progressively reduce the number of initia values (or starting points) 
%                       until we reach the full (final) optimization stage. Concretely, we run the CRL algorithm for a  
%                       small numer of steps in each pre-judgement stage to keep fewer and fewer starting points. 
%                       (User should provide both of its fields or leave whole the struct empty (so that a default struct will be constructed) 
%       -> NumsStarts: A decreasing row vector (2:end, excluding the 1st elem) containing the numbers of
%                      initial values used for each of the pre-judgement
%                      stages. The first element NumsStarts[1] will be not be used at all, and so we often set it to -1. 
%                      Also, the very last entry determines the number of initial values remaining for the final 
%                      optimization stage. [Some explanation: The first stage will keep NumsStarts[2] many starting values, and so 
%                      NumsStarts[1] will not be used in the program and can be set to any values (but do not set it to [] which will be auto removed). ]
%       -> ItersForPrejudgement: a row vector of length numel(NumsStarts) - 1.
%                                the ith entry gives the number of CRL iteration steps in the ith (pre-judgement) stage.
%                                Note: If not enough elements are provided by the user, we will repeat the last value to fill in the whole array
%   convCtrl: Check Func_CRL_singleinit for details
%   linesearchCtrl: Check Func_CRL_singleinit for details
% Output arguments:
%   B_est : Coefficient matrix estimate of size p x m.
%   alpha_est: Intercept estimate of size intrcptNo x 1.
%   S_est : Loading matrix of size p x r.
%   V_est : Rotation matrix of size m x r.
%   F_est : Binary membership matrix p x q.
%   mu_est : Cluster centroid matrix q x r. 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
debug = 0; %2; %  1; %
if ~contains(lossType, 'tracemodel')
    [n, m] = size(Y); p = size(X, 2); 
    intrcptNo = m; % num of components in the intercept column
else
    if size(X, 3) > 1
        % The most general trace model: y_i (scalar) ~ < X(:, :, i), B > + alpha_i
        [n] = size(Y, 1); 
        p = size(X, 1); m = size(X, 2); % B is of size pxm
        intrcptNo = n; % num of components in the intercept column
    else
        % if size(X, 3) == 1, i.e., a two-dim design matrix, 
        % we assume a "sample-based trace model": y_i (scalar) ~ < X(i, :), B(i, :) >  + alpha_i
        [n] = size(Y, 1); p = size(X, 1); m = size(X, 2); % B is of size pxm (or nxp in this special case)
        intrcptNo = n;
    end  
end

% Set the initial parameters
defaultPrejudgeIters = 3; %  Default number of iterations in each pre-judgement stage
defaultNumStarts = 23; %22; %  % Default number of initial points
if isfield(init, 'F') && ~isempty(init.F) && isfield(init, 'mu') && ~isempty(init.mu) && isfield(init, 'V') && ~isempty(init.V)
    % Meaningful initial values are provided by the user
    F_inits = init.F; mu_inits = init.mu; V_inits = init.V; 
    numStarts = size(F_inits, 3); % the number of points in inits provided by user
    if estimateIntercept == 1
        alpha_inits = init.alpha;
    else
        alpha_inits = zeros(intrcptNo, numStarts);
    end
    if isempty(multi_start_params) % no multi-stage parameter provided by the user
        if numStarts <= 1 % If there is only 1 initial value, perform no prejudgement. Directly go to the final stage (full opt.)
            multi_start_params.NumsStarts = []; 
            multi_start_params.ItersForPrejudgement = [];
        else
            multi_start_params.NumsStarts = [-1, max(1, floor(numStarts/2)), 1];
            multi_start_params.ItersForPrejudgement = [defaultPrejudgeIters, defaultPrejudgeIters];
        end
    else
        tempErrEntries = multi_start_params.NumsStarts > numStarts;
        if any(tempErrEntries) 
            warning('The elements of multi_start_params.NumsStarts (to be kept after each stage) is larger than the total number of initial vales provided by inits. Shrink it.')
            multi_start_params.NumsStarts(tempErrEntries) = numStarts;
        end
    end
else
    % No specific initial values are provided in init
    if isempty(numInits) % no number of initial values provided either
        % Use the default starting points
        numStarts = defaultNumStarts;
        numInits.rrr = unique([r, q, m]); %[q, m]; %  
        numInits.rand = max(0, numStarts - numel(numInits.rrr)); % 22 - 2 = 20  % 23 - 3 = 20 
    else
        if ~isfield(numInits, 'rrr') || ( isfield(numInits, 'rrr') && isempty(numInits.rrr) ) 
            % User did not specify 'rrr', or specify it as []
            numInits.rrr = unique([r, q, m]); %[q, m]; %
        else
            % User did specify 'rrr' at a nontrivial value
            if numInits.rrr == -1 % In this case, NO RRR estimates will be used as starting values
                numInits.rrr = []; % THIS EMPTY MEANS NO RRR ESTIMATES (while or input argument = [], use the default ranks)
            end
        end
        if ~isfield(numInits, 'rand') || ( isfield(numInits, 'rand') && isempty(numInits.rand) )
            % User did not specify 'rand', or specify it as []
            numInits.rand = max(0, defaultNumStarts - numel(numInits.rrr)); % 22 - 2 = 20 % 23 - 3 = 20 % 
        else % User did specify 'rand' at a nontrivial value            
            if numInits.rand < 0
                warning('Invalid numInits.rand');
                numInits.rand = 0;
            end
        end
        if isempty(numInits.rrr ) && numInits.rand == 0
        % In this case, init is empty, so is numInits.rrr, and numInits.rand == 0
        % So we should give at least one initial value to start the algorithm
            numInits.rand = 1;
        end

        numStarts = numel(numInits.rrr) + numInits.rand; 
    end
    if isempty(multi_start_params) % use the default
        multi_start_params.NumsStarts = [-1, max(1, floor(numStarts/2)), 1];
        multi_start_params.ItersForPrejudgement = [defaultPrejudgeIters, defaultPrejudgeIters];
    end    
    [B_inits, alpha_inits, S_inits, V_inits, F_inits, mu_inits]...
        = Func_GeneInits(X, Y, q, r, lossType, numInits, estimateIntercept);
end

NumsStageStarts = multi_start_params.NumsStarts;
ItersForPrejudgement = multi_start_params.ItersForPrejudgement;
numStages = max([numel(NumsStageStarts), 1]);
if numel(ItersForPrejudgement) < numStages - 1 
    % Fill in the remaining spots using the last entry of
    % ItersForPrejudgement provided by the user
    tempLengthIters = numel(ItersForPrejudgement);
    ItersForPrejudgement = [ItersForPrejudgement, ... 
        repmat( ItersForPrejudgement(end), [1, numStages - 1 - tempLengthIters] ) ];
end

currentInits.F = F_inits; currentInits.mu = mu_inits; currentInits.V = V_inits;
currentInits.alpha = alpha_inits;

numStarts = size(F_inits, 3);
if debug > 0, disp( [ 'Number of initial values is ', num2str(num2str( numStarts)), '; numInits.rrr=', num2str(numInits.rrr ), ', numInits.rand=', num2str(numInits.rand) ] ), end
if any(NumsStageStarts > numStarts)
    if debug > 0, disp('Some starting/remaining numbers are larger than the provided number of initial values! Shrink it'), end
    NumsStageStarts = min(NumsStageStarts, numStarts);
end
%-----------------Algorithm-----------------%
for stageInd = 1:numStages
    if stageInd == numStages 
        % Final stage: perform full optimziation in depth
        if debug > 0
            disp('++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++');
            disp('........................Final stage optimization..............................');
            disp('+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++');
        end
        calibRequired = 2; %1; %  0; %    0: no calibration; 1: perform calib only after the outer loop ends; 2: perform calib after each inner loop ends (at the end of every outer loop iteration and after the outer loop ends)
        numRemaining = 1;
        thisConvCtrl = convCtrl;
        [F_est, mu_est, V_est, alpha_est ] = ...
            Func_CRL_multipleinits_aux(Y, X, q, r, estimateIntercept, lossType,...
            currentInits, numRemaining, thisConvCtrl, calibRequired, linesearchCtrl);
        S_est = F_est * mu_est; 
        B_est = S_est * V_est';
    else
        % Pre-judgement stage: Reduce the number of initials by running CRL
        % for a small number of iterations
        calibRequired = 1; %0; %2; %   to save time in pre-judgement
        numRemaining = NumsStageStarts(stageInd + 1); 
        if debug > 0
            disp('++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++');
            disp(['............Pre-judgement stage: ' ,num2str(stageInd), ', with ', num2str( numStarts ), ' initials to obtain ' , num2str(numRemaining) , ' initials for next stage.']);
            disp('++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++');
        end
        itersForPrejudgement = ItersForPrejudgement(stageInd);
        thisConvCtrl = struct('outer', struct('maxiter', itersForPrejudgement));
        [ F_initsRem, mu_initsRem, V_initsRem, alpha_initsRem ] = ...
            Func_CRL_multipleinits_aux(Y, X, q, r, estimateIntercept, lossType,...
            currentInits, numRemaining, thisConvCtrl, calibRequired, linesearchCtrl);
        currentInits.F = F_initsRem; currentInits.mu = mu_initsRem; currentInits.V = V_initsRem;
        currentInits.alpha = alpha_initsRem;
        numStarts = numRemaining;
    end
end
end




function [ F_initsRem, mu_initsRem, V_initsRem, alpha_initsRem ] = ...
    Func_CRL_multipleinits_aux(Y, X, q, r, estimateIntercept, lossType,...
            inits, numRemaining, thisConvCtrl, calibRequired, linesearchCtrl)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Pick "numRemaining" many initial values by running CRL for a certain 
% amount of time based on the function value. 
% inits MUST have the following fields: F, mu, V (and alpha if
% estimateIntercept)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
debug = 0; 
if ~contains(lossType, 'tracemodel')
    [n, m] = size(Y); p = size(X, 2); 
    intrcptNo = m; % num of components in the intercept column
else
    if size(X, 3) > 1
        % The most general trace model: y_i (scalar) ~ < X(:, :, i), B > + alpha_i
        [n] = size(Y, 1); 
        p = size(X, 1); m = size(X, 2); % B is of size pxm
        intrcptNo = n; % num of components in the intercept column
    else
        % if size(X, 3) == 1, i.e., a two-dim design matrix, 
        % we assume a "sample-based trace model": y_i (scalar) ~ < X(i, :), B(i, :) >  + alpha_i
        [n] = size(Y, 1); p = size(X, 1); m = size(X, 2); % B is of size pxm (or nxp in this special case)
        intrcptNo = n;
    end  
end
% Obraining initial points
V_inits = inits.V; F_inits = inits.F; mu_inits = inits.mu; 
numStarts = size(F_inits, 3);
if estimateIntercept
    alpha_inits = inits.alpha;
else
    alpha_inits = zeros(intrcptNo, numStarts);
end

% Validating if sufficiently many initial points are present 
if numRemaining < 1 || numStarts < 1 % || numStarts < numRemaining
    error('The # of starting/remaining points must be at least 1')
end

fvals = -ones(numStarts, 1);
for initInd = 1:numStarts
    % Get the "initInd"-th initial value
    init.F = F_inits(:, :, initInd); init.mu = mu_inits(:, :, initInd); 
    init.V = V_inits(:, :, initInd); init.alpha = alpha_inits(:, initInd); 
    % To initialize the alg, only (F, mu, V) and alpha(if model includes intercept) are needed
    
    % Run the CRL algorithm with the single given init (with the convergence controlled by thisConvCtrl)
    [B_est, alpha_est, ~, V_est, F_est, mu_est]  = ...
        Func_CRL_singleinit(X, Y, q, r, init, lossType, estimateIntercept, thisConvCtrl, calibRequired, linesearchCtrl); 
	if debug > 0, disp(['CRL_multi-inits_aux initInd=', num2str(initInd), '. Change in F? ', num2str(max(max(abs(F_est - init.F))))]); end
    % Replace the intial value of each parameter by its current iterate 
    F_inits(:, :, initInd) = F_est; 
    mu_inits(:, :, initInd) = mu_est;
    V_inits(:, :, initInd) = V_est;  
    alpha_inits(:, initInd) = alpha_est;
    
    % Calculate and record the functiona value
%     [Theta] = Func_Eval_Theta_SysComp(X, B_est, alpha_est, lossType); % For ordinary models, Theta = X * B_est + 1 * alpha_est';
%     fvals(initInd) = Func_Loss_Grad(Y, Theta, lossType, 0); %0: loss value only
    fvals(initInd) = Func_Loss_Grad_Balpha(Y, X, B_est, alpha_est, lossType, 0); %0: loss value only

end

% Pick the best numRemaining inits based on their function values
[~, sort_idx] = sort(fvals, 'ascend');
remInds = sort_idx(1:numRemaining);
F_initsRem = F_inits(:, :, remInds);
mu_initsRem = mu_inits(:, :, remInds);
V_initsRem = V_inits(:, :, remInds);
alpha_initsRem = alpha_inits(:, remInds);
end
