function [B_est, alpha_est, S_est, V_est, F_est, mu_est]  = ...
    Func_CRL_singleinit(X, Y, q, r, init, lossType, estimateIntercept, convCtrl, calibRequired, linesearchCtrl)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This function solves supervised CRL problem for a given single 
% initialization, i.e., 
% min_{S, V, alpha} l_0(1*alpha' + X*S*V^T; Y) s.t.  ||S||_{2,C} <= q , V^T*V=I.
%
% Input arguments:
%   X: the design matrix/matrices (no intercept column for ordinary matrix models, but may contain it for sample-based trace models)
%   Y: the response matrix (Y ~ XB + 1 * alpha'), 
%      or the response vector for trace models, either the general one 
%      y_i (scalar) ~ < X(:, :, i), B > + alpha_i or a "sample-based trace model": 
%      y_i (scalar) ~ < X(i, :), B(i, :) >  + alpha_i
%   q: the cluster size of interest.
%   r: the rank value of interest.
%   init: Structure array containing initial values of F, mu, V and B. (and
%         of alpha if estimateIntercept = 1)
%         B is only used if F, mu and V are not all provided.
%       -> init. B: Initial coefficeint matrix.
%       -> init. V: Initial rotation matrix.
%       -> init. F: Initial binary membership matrix.
%       -> init. mu: Initial cluster centroid matrix.       
%       -> init. alpha: Initial intercept column vector.
%   losstype: Could be 'gaussian', 'bernoulli', 'poisson', a robust loss, 
%             or any other loss of interest (not all implemented yet). 
%             Check Func_Loss_Grad for more details.
%   estimateIntercept: (Default: 1)
%       0: No intercept estimation (return zeros).
%       1: The intercept alpha will be simultaneously estimated.    
%   convCtrl: A structure array for convergence control.
%       -> convCtrl.outer: A structure array for convergence control of the outer 
%                          loop. Default values provided if all fields are empty/vacant. 
%                          (Default: maxiter = 20; B_tol = 1e-6) 
%           => convCtrl.maxiter: maximum number of iterations allowed.
%           => convCtrl.B_tol: estimate error tolerance. 
%       -> convCtrl.inner: A structure array for convergence control inside 
%                          CRL algorithm inner loop. (Check Func_unsuperv for more details.)
%           => convCtrl.inner.maxiter: maximum number of iterations allowed.
%           => convCtrl.inner.B_tol: estimate error tolerance. 
%           => convCtrl.inner.func_tol: objective function value tolerance.
%           => convCtrl.inner.F_tol: Membership matrix convergence tolerance.
%                                    0 if not required, 1 if required.
%   calibRequired: Controlling post calibration on CRL estimate (default: 1) 
%                  2: perform the calibration after each inner loop ends;
%                  1: perform it only after the outer loop ends; 0: no calibration
%   linesearchCtrl: A structure array for line search control. If not
%                   provided, use the default constructed below in the code 
%       -> linesearchCtrl.rho_redfactor: multiplicative reduction factor in performing line search
%       -> linesearchCtrl.numRhosInTrial: search times
%       -> linesearchCtrl.linesearchDir: search direction, either in an
%           adaptive manner (use the previous learning rate and search in both directions)
%           or backward manner (use the theorectical (conservative) choice to backtrack)
%
% Output arguments:
%   B_est : Coefficient matrix estimate of size p x m.
%   alpha_est: Intercept estimate of size intrcptNo x 1.
%   S_est : Loading matrix of size p x r.
%   V_est : Rotation matrix of size m x r.
%   F_est : Binary membership matrix p x q.
%   mu_est : Cluster centroid matrix q x r. 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
debug = 0; %1; %2; %    %Show some debugging information
if ~contains(lossType, 'tracemodel')
    [n, m] = size(Y); p = size(X, 2); 
    intrcptNo = m; % num of components in the intercept column
else
    if size(X, 3) > 1
        % The most general trace model: y_i (scalar) ~ < X(:, :, i), B > + alpha_i
        [n] = size(Y, 1); 
        p = size(X, 1); m = size(X, 2); % B is of size pxm
        intrcptNo = n; % num of components in the intercept column
    else
        % if size(X, 3) == 1, i.e., a two-dim design matrix, 
        % we assume a "sample-based trace model": y_i (scalar) ~ < X(i, :), B(i, :) >  + alpha_i
        [n] = size(Y, 1); p = size(X, 1); m = size(X, 2); % B is of size pxm (or nxp in this special case)
        intrcptNo = n;
    end  
end


% Controlling the requirement of calibration
if ~exist('calibRequired', 'var') || isempty(calibRequired)
    calibRequired = 1; % 0: no further calibration; 1: perform calib only after the outer loop ends; 2: perform calib after each inner loop ends (at the end of every outer loop iteration and after the outer loop ends)
end

% Setting default values for convergence control parameters if user hasn't
% provided any control parameter
deflt_convctrl_outer = struct('maxiter', 20, 'B_tol', 1e-6);
if ~isfield(convCtrl, 'outer') || isempty(convCtrl.outer)
    convCtrl.outer = deflt_convctrl_outer;
end
if ~isfield(convCtrl, 'inner')
    innerConvCtrl = [];
else
    innerConvCtrl = convCtrl.inner; % If it's empty or values are given
end

if ~exist('estimateIntercept', 'var') || isempty(estimateIntercept)
    estimateIntercept = 1;
end

% Preprocessing for convergence control
convstatus = zeros(1, numel(fieldnames(convCtrl.outer))+1); % to obtain convergence condition(s)
convcriteNo = 1; % the # of convergence criteria needs to be satisfied

% Set the starting point for all necessary unknowns before running the loop
if isfield(init, 'F') && isfield(init, 'mu') && isfield(init, 'V') 
    F_cur = init.F; mu_cur = init.mu; V_cur = init.V; 
    S_cur = F_cur * mu_cur; B_cur = S_cur * V_cur'; 
    if estimateIntercept && isfield(init, 'alpha') 
        alpha_cur = init.alpha;
    else
        alpha_cur = zeros(intrcptNo, 1);
    end
elseif ( ~isfield(init, 'F') || ~isfield(init, 'mu') || ~isfield(init, 'V') ) && isfield(init, 'B')
    % use the given B to generate the initial values
	[B_cur, alpha_cur, S_cur, V_cur, F_cur, mu_cur] = Func_RoughClusterB_aux(X, Y, ...
            init.B, q, r, lossType, estimateIntercept, []);  % []: use the default convCtrl in calling calibration 
else 
    % If no initial values are provided, we will use the rank-q RRR solution for initialization 
    [B_cur, alpha_cur, S_cur, V_cur, F_cur, mu_cur]...
        = Func_GeneInits(X, Y, q, r, lossType, ...
            struct('rrr', q, 'rand', 0), estimateIntercept); %  a single rank-q solution from RRR for numInits
end

%---------CRL algorithm-----------%
% Specify the inverse stepsize parameter
rhofactor = 1; % 1: conservative choice % to adjust the conservative theoretical lower bound of rho so that a larger stepsize will be used
[rho_ref] = Func_RhoRef_InvStepsize(X, lossType, estimateIntercept, rhofactor);
rho_min = 1e-5;
% default parameter values for line search
deflt_rho_redfactor = 0.5; % 0.25
deflt_numRhosInTrial = 5;  % 5
deflt_linesearchDir = 'backward'; %'adaptive' %  
if ~exist('linesearchCtrl', 'var')
    rho_redfactor = deflt_rho_redfactor;
    numRhosInTrial = deflt_numRhosInTrial;
    linesearchDir = deflt_linesearchDir;
    if debug > 1, disp('A default line search criterion is used'); end
else
    if ~isfield(linesearchCtrl, 'linesearchDir') || isempty(linesearchCtrl.linesearchDir)
        linesearchDir = deflt_linesearchDir; 
    else
        linesearchDir = linesearchCtrl.linesearchDir; 
    end
    if ~isfield(linesearchCtrl, 'rho_redfactor') || isempty(linesearchCtrl.rho_redfactor)
        rho_redfactor = deflt_rho_redfactor; 
    else
        rho_redfactor = linesearchCtrl.rho_redfactor;
    end
    if ~isfield(linesearchCtrl, 'numRhosInTrial') || isempty(linesearchCtrl.numRhosInTrial)
        numRhosInTrial = deflt_numRhosInTrial;
    else
        numRhosInTrial = linesearchCtrl.numRhosInTrial;
    end
end

%----------- Start the outer loop  ------------------%
if debug > 1, disp('--------------- CRL main/outer loop starts (with line search)---------------'); end
k = 0;
rho_prev = rho_ref;
while 1
    k = k + 1;
    if debug > 1, disp(['*************** outer loop index k = ', num2str(k), ' *******************']); end
    % Pre-calc some quantities for linerization
    [XtG, Theta, GtOnes] = Func_GradB(X, Y, lossType, estimateIntercept, B_cur, alpha_cur);
    
    unsupvinit.F = F_cur;  unsupvinit.mu = mu_cur; unsupvinit.V = V_cur;    
    if strcmp(linesearchDir, 'adaptive')
        rho_cur = rho_prev; % rho_ref; % Warm start the line search, always from the given initial rho
    else
        rho_cur =  rho_ref; % Cold start the line search, always from the given initial rho
    end
    
    linesearchInd = 0;
    critval_prev = -1;
    newvalassigned = false;
    while 1
        linesearchInd = linesearchInd + 1;
        if debug >= 2 
            disp(['Line search iteration number: ', num2str(linesearchInd)]); 
        end
        % 1. Use the basic optimization procedure and the current rho to get the some trial values of the parameters to be estimated
        [alpha_trial, B_trial, S_trial, V_trial, F_trial, mu_trial] = ...
            parameterUpdate_local(rho_cur, B_cur, alpha_cur, q, XtG, ...
                GtOnes, estimateIntercept, unsupvinit, innerConvCtrl);
        % 2. Pre-calculation of some quantities to be used for line search 
        [Theta_trial] = Func_Eval_Theta_SysComp(X, B_trial, alpha_trial, lossType); % ordinary model: Theta_trial = X * B_trial + alpha_trial';
        % Theta_trial needs to be calculated outside linesearchCrit_local
        % because it is required to calculate obj_err and gives us
        % flexibilty to add it as a convergence criterion in future.
        critval_cur = linesearchCrit_local(Y, B_trial, B_cur, alpha_trial, alpha_cur, Theta_trial, Theta, rho_cur, lossType, debug); % 0 or 1, based on trial values!!!
        if debug >= 2, disp(['Is the line search criterion met for the current rho(=', num2str(rho_cur), ')? --- ', num2str(critval_cur)]), end
%         if linesearchInd == 1 && critval_cur ~= 1
%             error('rho factor should be 1 to start the backtracking on rho!')
%         end
        if critval_cur == 1 % && rho_prev > rho_cur % Update 
            [alpha_new, B_new, S_new, V_new, F_new, mu_new] = deal(alpha_trial, B_trial, S_trial, V_trial, F_trial, mu_trial);
            Theta_new = Theta_trial;
            if ~newvalassigned,  newvalassigned = true; end
        end
        % 3. modify rho by line search
        if debug >= 2, disp(['Curr RHO =', num2str(rho_cur),';  Prev RHO = ', num2str(rho_prev)]); end
        if linesearchInd >= numRhosInTrial || (critval_prev == 1 && critval_cur == 0) || (critval_prev == 0 && critval_cur == 1)
            % if linesearchInd >= numRhosInTrial || (critval_prev == 1 && 
            % critval_cur == 0) in the same iteration, rho_prev will be
            % rho_cur while it should be rho_prev (For backward search, the last condition will never be met)
            if (critval_prev == 0 && critval_cur == 1) || (critval_prev == 1 && critval_cur == 1)
                if debug >= 2, disp(['Final rho is ---', num2str(rho_cur),', found in ', num2str(linesearchInd), 'th Line search step. Prev-rho=', num2str(rho_prev), ', Cur-rho=', num2str(rho_cur)]); end
                rho_prev = rho_cur;
            elseif (critval_prev == 1 && critval_cur == 0)
                if debug >= 2, disp(['Final rho is ---', num2str(rho_prev),', found in ', num2str(linesearchInd-1), 'th Line search step. Prev-rho=', num2str(rho_prev), ', Cur-rho=', num2str(rho_cur)]); end
                rho_prev = rho_prev;
            elseif (critval_prev == 0 && critval_cur == 0)
                if debug >= 2, disp(['Final rho is ---', num2str(rho_cur),', found in ', num2str(linesearchInd), 'th Line search step. Prev-rho=', num2str(rho_prev), ', Cur-rho=', num2str(rho_cur)]); end
                rho_prev = rho_cur;
                if debug >= 1
                    if isfield(convCtrl.outer, 'maxiter') && ~isempty(convCtrl.outer.maxiter) && k >= convCtrl.outer.maxiter
                        warning('This final rho is too small for the last iteration of outerloop. Increase the number of searches, otherwise the algorithm is divergent');
                    else
                        warning('This final rho is too small. Try to increase the number of searches, otherwise the algorithm may diverge');
                    end
                end
            end
            if ~newvalassigned % linesearchInd >= numRhosInTrial is a necessary condition
                [alpha_new, B_new, S_new, V_new, F_new, mu_new] = deal(alpha_trial, B_trial, S_trial, V_trial, F_trial, mu_trial);
                Theta_new = Theta_trial;
            end
            break;
        else % linesearchInd < max, && [ (critval_prev == 0 && critval_cur == 0) || (critval_prev == 1 && critval_cur == 1) ]
            if critval_cur == 1 % (prev, cur) = (1, 1) (or (-1, 1) in the 1st iteration)
                rho_prev = rho_cur;
                rho_cur = max(rho_min, min(rho_cur * rho_redfactor, rho_ref));
                critval_prev = critval_cur;
            else % (prev, cur) = (0, 0) (or (-1, 0) in the 1st iteration)
                 % 'backward' will not reach this case
                if strcmp(linesearchDir, 'backward'), error('Wrong entrance into increasing rho'), end
                rho_prev = rho_cur;
                rho_cur = min(rho_cur / rho_redfactor, rho_ref);
                critval_prev = critval_cur;
            end
        end
    end
    
    %%% Calibration (if inner loop terminated earlier/ not fully convergent)
    % The default convCtrl constructed in Func_Calibration is used, with alpha_new, mu_new and V_new used for initialization    
    if calibRequired >= 2
        [B_calib, alpha_calib, ~, ~, ~, ~] ...
            = Func_Calibration(X, Y, lossType, estimateIntercept,...
                [], r, F_new, alpha_new, mu_new, V_new);
%         Func_Loss_Grad_Balpha(Y, X, B_calib, alpha_calib, lossType, 0)
%         [B_calib, alpha_calib, ~, ~, ~, ~] ...
%             = Func_Calibration(X, Y, lossType, estimateIntercept,...
%                 [], r, F_new);
%         Func_Loss_Grad_Balpha(Y, X, B_calib, alpha_calib, lossType, 0)
        
        B_new = B_calib;
        alpha_new = alpha_calib;
    end    
    %-----------------stopping criterion-------------%
    B_err = max(max(abs(B_new - B_cur))); 
    obj_cur = Func_Loss_Grad_Balpha(Y, X, B_cur, alpha_cur, lossType, 0); % Func_Loss_Grad(Y, Theta, lossType, 0) if no calib
    obj_new = Func_Loss_Grad_Balpha(Y, X, B_new, alpha_new, lossType, 0); % Func_Loss_Grad(Y, Theta_new, lossType, 0) if no calib
    obj_change = abs(obj_new-obj_cur);
    % Checking for convergence
    if isfield(convCtrl.outer, 'maxiter') && ~isempty(convCtrl.outer.maxiter) && k >= convCtrl.outer.maxiter
        convstatus(1) = 1;
    end
    if isfield(convCtrl.outer, 'B_tol') && ~isempty(convCtrl.outer.B_tol) && B_err < convCtrl.outer.B_tol
        convstatus(2) = 1;
    end
    
    %------------Print values-------------------%
    if debug > 0
        F_change = max(max(abs(F_new - F_cur)));
        if debug > 1
            disp(['Outer loop index=',num2str(k), '. Change in B: ', num2str(B_err),', obj fnval= ', num2str(obj_new), ', obj change= ', num2str(obj_change),  '. Change in F: ', num2str(F_change)]); 
        else
            disp(['Outer loop index=',num2str(k), '. Change in F: ', num2str(F_change)]); 
        end
    end
    
    % The algorithm converges if convcriteNo number of conditions are satisfied.
    if nnz(convstatus)>=convcriteNo
        break;
    end
    
    B_cur = B_new; alpha_cur = alpha_new; S_cur = S_new; V_cur = V_new; F_cur = F_new; mu_cur = mu_new; 
end

if calibRequired >= 1
    [B_calib, alpha_calib, S_calib, V_calib, F_calib, mu_calib] = Func_Calibration(X, Y, lossType, estimateIntercept, [], r, F_new, alpha_new, mu_new, V_new);
%     disp('Calib end')
%     Func_Loss_Grad_Balpha(Y, X, B_calib, alpha_calib, lossType, 0)
%     [B_calib, alpha_calib, ~, ~, ~, ~] ...
%         = Func_Calibration(X, Y, lossType, estimateIntercept,...
%             [], r, F_new);
%     Func_Loss_Grad_Balpha(Y, X, B_calib, alpha_calib, lossType, 0)
    if debug > 0
        disp(['B change after calibration: ', num2str( max(max(abs(B_new - B_calib))) )]) % it seems that for regression, this often showing mild differences
        disp(['f-value after calib - original: ', num2str( Func_Loss_Grad_Balpha(Y, X, B_calib, alpha_calib, lossType, 0) - Func_Loss_Grad_Balpha(Y, X, B_new, alpha_new, lossType, 0) )])  
    end
    
    B_est = B_calib; alpha_est = alpha_calib; S_est = S_calib; V_est = V_calib; F_est = F_calib; mu_est = mu_calib; 
else
    B_est = B_new; alpha_est = alpha_new; S_est = S_new; V_est = V_new; F_est = F_new; mu_est = mu_new; 
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if debug > 1
    disp(['CRL outer loop convergence condition: [', num2str(convstatus), ']']);
    disp('--------------- Main loop ends here ---------------'); 
end
if debug > 0, disp(['*************** CRL_singleinit is over, with the final outer loop iters=', num2str(k), ' *******************']), end
end


function [critval] = linesearchCrit_local(Y, B_trial, B_old, alpha_trial, alpha_old, Theta_trial, Theta_old, rho_trial, lossType, debug)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Check if the line search condition is met
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
[F_old, G_old] = Func_Loss_Grad(Y, Theta_old, lossType); 
[F_trial, ~] = Func_Loss_Grad(Y, Theta_trial, lossType, 0); 
% 1: loss(Theta_trial) <= loss(Theta_old) + <G( Theta_old) .* (Theta_trial - Theta_old)> ...
%     + rho_trial * ((B_trial - B_old))_F.^2/2 +  rho_trial * ((alpha_trial - alpha_old))_F.^2/2

critval_difference = F_old + sum(sum(G_old .* (Theta_trial - Theta_old)))...
                        + (rho_trial * norm(B_trial - B_old, 'fro').^2/2) + (rho_trial * norm(alpha_trial - alpha_old).^2/2) ...
                     - F_trial;
if critval_difference > -(1e-6) % 0 % 
    critval = 1;
else
    critval = 0;
end

if debug >= 2
    disp (['G_',num2str(rho_trial),'(B_trial,B_old)-F(B_trial) = ', num2str(...
        F_old + sum(sum(G_old .* (Theta_trial - Theta_old)))...
            + (rho_trial * norm(B_trial - B_old, 'fro').^2/2) + (rho_trial * norm(alpha_trial - alpha_old).^2/2) ...
            - F_trial)]);
end
end

function [alpha_new, B_new, S_new, V_new, F_new, mu_new] = ...
    parameterUpdate_local(rho_cur, B_cur, alpha_cur, q, XtG, GtOnes, estimateIntercept, unsupvinit, innerConvCtrl)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Local function to calculate the inner loop
% 1. Update alpha with new gradient
% 2. Update B, S, V, F, mu in the inner loop of CRL
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 1. Alpha update
if estimateIntercept == 1
    alpha_new = alpha_cur - GtOnes / rho_cur; % i.e., alpha_new' = alpha_cur' - grad_alpha / rho 
else
    alpha_new = zeros(size(alpha_cur));
end        
% Update pseudo Y for innter loop the coefficient matrix B   
Y_tilde = B_cur - XtG / rho_cur;
% 2. Inner loop, to solve an unsupervised CRL
[B_new, S_new, V_new, F_new, mu_new] = Func_CRL_unsuperv(Y_tilde, q, unsupvinit, innerConvCtrl);
end
