 function [B_est, S_est, V_est, F_est, mu_est] = ...
    Func_CRL_unsuperv(Y, q, init, convCtrl) 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Unsupervised CRL
% This function solves the CRL in the unsupervised setup with a quadratic 
% loss and NO intercept, i.e., 
%      min_{S, V} || Y - S*V^T||_F^2/2  s.t.  ||S||_{2,C} <= q , V^T V=I.
% No calibration is performed. We just run an iterative algorithm which
% stops whenever the stopping criterion is met.
%
% Input arguments:
%   Y : Response/ pseudo-response matrix for unsupervised CRL problem.
%   q: Cluster number.
%   init: Structure array containing the initial values of F, mu and V. 
%            (Following fields MUST be provided for this algorithm.)
%       -> init.V: Initial rotation matrix.
%       -> init.F: Initial binary membership matrix.
%       -> init.mu: Initial cluster centroid matrix.
%   convCtrl: A structure array for convergence control. Default values
%             provided only if all fields are empty/vacant. (**Ask  professor if the word is good)         
%             (Default: maxiter = 20; B_tol = 1e-4; func_tol = 1e-4; F_tol = 1)     
%       -> convCtrl.maxiter: maximum number of iterations allowed.
%       -> convCtrl.B_tol: estimate convergence tolerance. 
%       -> convCtrl.func_tol: objective function value tolerance.
%       -> convCtrl.F_tol: Membership matrix convergence tolerance.
%                          (0 if not desired, 1 if desired.)
%
% Output arguments:
%   B_est : Coefficient matrix estimate of size p x m.
%   S_est : Loading matrix of size p x r.
%   V_est : Rotation matrix of size m x r.
%   F_est : Binary membership matrix p x q.
%   mu_est : Cluster centroid matrix q x r. 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
debug = 0;
% Setting default values for convergence control parameters if user hasn't
% provided any control parameter
deflt_convCtrl = struct('maxiter', 20, 'B_tol', 1e-4, 'func_tol', 1e-4, 'F_tol', 1);
if ~exist('convCtrl', 'var') || isempty(convCtrl)
    convCtrl = deflt_convCtrl;
end

% Preprocessing for convergence control
F_stab = 0; F_stab_limit = 3; % Used for F-stability based convergence judgement
convstatus = zeros(1, numel(fieldnames(convCtrl))); % to obtain convergence condition(s)
convcriteNo = 1; % the # of convergence criteria needs to be satisfied


% Preprocessing the given initial parameters to feed into the algorithm
F_cur = init.F; mu_cur = init.mu;  V_cur = init.V; 
S_cur = F_cur * mu_cur; B_cur = S_cur * V_cur'; 

% Controlling k-means algorithm
clusteringway = 'kmeans'; % 'kmeans' % 'kmeanspp' % 'both'%  
% global kmeansinnerloop totalinnerloop % for debugging

%-----------Unsupervised CRL algorithm---------------%
t = 0;
if debug >= 1, disp('--------------Inner loop start (BCD)--------------'); end
while 1
    t = t + 1;
    %----------V-optimization----------------%
    W = Y' * S_cur;
    [U_w, ~, V_w] = svd(W, 'econ');
    V_new = U_w * V_w';
    L = Y * V_new;
    %----------S-optimization-------------------
    if strcmp(clusteringway,'kmeans') || strcmp(clusteringway,'both')
        [ idx_mat_kmeans, center_mat_kmeans ] = Func_kmeans_aux(L, q, mu_cur);   % k-means with initial center=mu.
        S_k = idx_mat_kmeans * center_mat_kmeans;
    end
    if strcmp(clusteringway,'kmeanspp') || strcmp(clusteringway,'both')
        [ idx_mat_kmeanspp, center_mat_kmeanspp ] = Func_kmeanspp_aux(L, q); % k-means++
        S_kpp = idx_mat_kmeanspp * center_mat_kmeanspp;
    end
    if strcmp(clusteringway,'both')
        tmp_L_error_kmeans = norm(L - S_k, 'fro');
        tmp_L_error_kmeanspp = norm(L - S_kpp, 'fro');
        if tmp_L_error_kmeans <= tmp_L_error_kmeanspp
            if debug
                if abs(tmp_L_error_kmeans - tmp_L_error_kmeanspp) < 1e-3
                    disp('Unspv CRL: K-means is comparable to K-means++')
                else
                    disp('Unspv CRL: K-means is better than K-means++')
                end
            end
            S_new = S_k;
            F_new = idx_mat_kmeans;
            mu_new = center_mat_kmeans;
%             kmeansinnerloop = kmeansinnerloop + 1; % debugging
        else
            if debug 
                disp('Unspv CRL: K-means++ is better than K-means')
            end
            S_new = S_kpp;
            F_new = idx_mat_kmeanspp;
            mu_new = center_mat_kmeanspp;
        end
    else
        if strcmp(clusteringway,'kmeans')
            S_new = S_k;
            F_new = idx_mat_kmeans;
            mu_new = center_mat_kmeans;
        end
        if strcmp(clusteringway,'kmeanspp')
            S_new = S_kpp;
            F_new = idx_mat_kmeanspp;
            mu_new = center_mat_kmeanspp;
        end
    end
%     totalinnerloop = totalinnerloop + 1; % debugging
    B_new = S_new * V_new';
    % Preprocessing for convergence checking
    B_err = max(max(abs(B_new - B_cur)));
    obj_err = abs( norm(Y - B_new, 'fro')^2/2 - norm(Y - B_cur, 'fro')^2/2 );
    if isequal(F_new, F_cur)
        F_stab = F_stab + 1;
    else
        F_stab = 0; 
    end
    % Checking for convergence
    if isfield(convCtrl, 'maxiter') && ~isempty(convCtrl.maxiter) && t >= convCtrl.maxiter
        convstatus(1) = 1;
    end
    if isfield(convCtrl, 'B_tol') && ~isempty(convCtrl.B_tol) && B_err < convCtrl.B_tol
        convstatus(2) = 1;
    end
    if isfield(convCtrl, 'func_tol') && ~isempty(convCtrl.func_tol) && obj_err < convCtrl.func_tol
        convstatus(3) = 1;
    end
    if isfield(convCtrl, 'F_tol') && ~isempty(convCtrl.func_tol) && convCtrl.F_tol && F_stab >= F_stab_limit  %isequal(F_new,F_cur) % or isequal
        convstatus(4) = 1;
    end
    % The algorithm converges if convcriteNo number of conditions are satisfied.
    if nnz(convstatus)>=convcriteNo % && convstatus(4) == 1 % Emphasizing on F stability
        break;
    end
    S_cur = S_new; V_cur = V_new; F_cur = F_new; mu_cur = mu_new; B_cur = B_new;
end
S_est = S_new; V_est = V_new; F_est = F_new; mu_est = mu_new; B_est = B_new;

if debug
%     if strcmp(clusteringway,'both') % to be corrected
%         disp(['total inner loop --- ', num2str(totalinnerloop)]);
%     end
    disp(['Unsupervised CRL BCD convergence condition: [', num2str(convstatus), '] with iter= ', num2str(t), ...
         ', B change= ', num2str(B_err), ', obj change= ', num2str(obj_err), ', F_stab= ', num2str(F_stab) ]);
    disp('--------------This inner loop ends--------------');
end
end
