function [B_calib, alpha_calib, S_calib, V_calib, F_calib, mu_calib] ...
    = Func_Calibration(X, Y, lossType, estimateIntercept, convCtrl, r, F,...
                        alpha, mu, V, B, S)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Post-calibration function. 
% This function does post calibration by running RRL on clustered design 
% matrix/matrices (X * F in the matrix model case, and F' * X(:, :, i) in the trace model case. 
% F is the binary membership matrix of clustered features. 
% 
% Notes: 1. For ordinary matrix models, the input X must not include the intercept column
%        2. The corresponding coefficient matrix B will not contain 
%           intercept estimate. We will provide a separate intercept
%           estimate alpha as a column vector.
%        3. Only F is required in the calibration. But mu, alpha and V, if 
%           provided, can be used to construct the initial point in calling 
%           RRL for calibration. 
%  
% Input arguments:
%   X: the design matrix/matrices (no intercept column for ordinary matrix models, but may contain it for sample-based trace models)
%   Y: the response matrix (Y ~ XB + 1 * alpha'), 
%      or the response vector for trace models, either the general one 
%      y_i (scalar) ~ < X(:, :, i), B > + alpha_i or a "sample-based trace model": 
%      y_i (scalar) ~ < X(i, :), B(i, :) >  + alpha_i
%   r: the rank value of interest.
%   losstype: Could be 'gaussian', 'bernoulli', 'poisson', a robust loss, 
%             or any other loss of interest (not all implemented yet). 
%             Check Func_Loss_Grad for more details.
%   estimateIntercept: (Default: 1)
%       0: No intercept estimation (return zeros).
%       1: The intercept alpha will be simultaneously estimated.      
%   convCtrl: A structure array for convergence control
%       -> convCtrl.maxiter: maximum number of iterations allowed.
%                            (Default: 50.)
%       -> convCtrl.tol: error tolerance. (Default: 1e-04.)
%   r : the rank upper bound of B.
%   F : The binary membership matrix of size p x q, associated with B.
%   alpha: The given intercept column.
%   mu : The cluster centroid matrix of size q x r, associated with B.
%   V : The rotation matrix of size m x r, associated with B.
%   B : The given coefficient matrix of size p x m.
%   S : The loading matrix of size p x r, associated with B.
%
% Output arguments:
%   B_calib : Calibrated coefficient matrix estimate of size p x m.
%   alpha_calib : Calibrated intercept estimates (column vector).
%   S_calib : calibrated loading matrix of size p x r.
%   V_calib : calibrated rotation matrix of size m x r.
%   F_calib : calibrated binary membership matrix of size p x q.
%   mu_calib : calibrated cluster centroid matrix of size q x r.
%
% Additional notes: 1. For matrix models, we will minimize l_0(1 * alpha' +
%                      Z * mu * V'; Y) with respect to C := mu * V' which
%                      has low rank, whre Z = X * F, and the RRL coefficient matrix is C. 
%                      For trace models with loss \sum_i l_0(alpha(i)  + < Z(:,:,i), mu * V' >; Y(i)),
%                      where Z is constructed from X and F, again, we let C := mu * V' subject to low rank
%                   2. The calibrated estimates are finally obtained by B =: F * C, S =: F * C * V. 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% B_ori = B; S_ori = S;  % might be used later. (Not required currently.)
F_calib = F; % membership matrix does not change
if r >= min([size(F, 1), size(F, 2), size(X, 2)]) && strcmp(lossType, 'gaussian-tracemodel') && size(X, 3) == 1 && isempty(convCtrl)
    % In this case, we have a sample-based trace regression model (no further low rank). 
    % Use the observations in each group to get the corresponding
    % coefficient estimate. [When convCtrl is specified, we run RRL according to the control params rather than the following]
    [B_calib, alpha_calib, C_calib, ~] = Func_FullmodelMLE(X, Y, estimateIntercept, lossType, F_calib);
    [~, ~, V_calib] = svd(C_calib, 'econ');
    mu_calib = C_calib * V_calib; S_calib = B_calib * V_calib;
    return;
end

% Form a new design (of size n x q in matrix models, or a series of new designs each of size q x m in trace models)
if ~contains(lossType, 'tracemodel')
    % ordinary matrix model Y ~ X * B + 1 * alpha' if intercept exists
    Z = X * F_calib;
else % In this case, we assume a trace model without intercept
    if estimateIntercept ~= 0, warning('The choice of rho in the trace model with an intercept is NOT implemented yet'), end
    if size(X, 3) > 1
        % The general trace model: y_i (scalar) ~ < X(:, :, i), B >
        Z = zeros(size(F_calib, 2), size(X, 2), size(X, 3));
        for i = 1:size(X, 3)
            Z(:, :, i) = F_calib' * X(:, :, i);
        end
    else
        % if size(X, 3) == 1, i.e., a two-dim design matrix, we assume a sample-based model: y_i (scalar) ~ < X(i, :), B(i, :) >
        Z = zeros(size(F_calib, 2), size(X, 2), size(Y, 1));
        for i = 1:size(Y, 1)
            X_aug_tmp = zeros(size(X)); X_aug_tmp(i, :) = X(i, :);
            Z(:, :, i) = F_calib' * X_aug_tmp;
        end
        % Now we changed the model formulation into the standard trace
        % model, with a 3-d array Z as the design
    end
end


if nargin == 7
    init = []; % Use the default initial vals constructed in Func_RRL.
    if strcmp(lossType, 'gaussian-tracemodel') && size(X, 3) == 1 && isempty(convCtrl)
        % In this case, we have a sample-based trace regression model with reduced rank
        % Use the full-rank estimate + PCA to give an initial estimate.
        % [When convCtrl is specified, we run RRL using its default rather than the following init]
        [~, ~, C_fr, ~] = Func_FullmodelMLE(X, Y, estimateIntercept, lossType, F_calib);
        [~, ~, V_fr] = svd(C_fr, 'econ');
        init.B = C_fr * V_fr(:, 1:r) * V_fr(:, 1:r)';
    end
else
    init.alpha = alpha; init.B = mu * V';
end

% Run RRL
if contains(lossType, 'tracemodel') %&& size(X, 3) > 1
    % Calculate a reference stepsize (to save a lot of computational cost)
    rho = Func_RhoRef_InvStepsize(X, lossType, estimateIntercept) * norm(F_calib, 2)^2 * .5; % add an additional factor 0.5 to alleviate the conserveness a bit
else
    rho = [];
end
% RRL on clustered features
[C, alpha_calib, mu_calib, V_calib] = Func_RRL(Z, Y, r, lossType, ...
    estimateIntercept, convCtrl, init, rho);
% Extracting coefficient and loading matrix.
B_calib = F_calib * C;
S_calib = F_calib * mu_calib;
end




