function [B_inits, alpha_inits, S_inits, V_inits, F_inits, mu_inits] = ...
    Func_GeneInits(X, Y, q, r, lossType, numInits, estimateIntercept)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This function generates initial points to start CRL algorithm. 
% Since intercepts are not subjected to any regularization, it is suggested
% to use column centered X and Y.
%
% Types of initials generated:
%   1. Feasible initials (V of size m x r and mu of size q x r) are
%      obtained from RRR estimates for given choice(s) of r.
%   2. Feasible initials (V of size m x r and mu of size q x r) are
%      obtained from random initials.
%
% Input arguments:
%   X: the design matrix/matrices (no intercept column for ordinary matrix models, but may contain it for sample-based trace models)
%   Y: the response matrix (Y ~ XB + 1 * alpha'), 
%      or the response vector for trace models, either the general one 
%      y_i (scalar) ~ < X(:, :, i), B > + alpha_i or a "sample-based trace model": 
%      y_i (scalar) ~ < X(i, :), B(i, :) >  + alpha_i
%   q: Desired maximum number of feature clusters.
%   r: Desired maximum rank of B.
%   losstype: Could be 'gaussian', 'bernoulli', 'poisson', a robust loss, 
%             or any other loss of interest (not all implemented yet). 
%             Check Func_Loss_Grad for more details.
%   numInits: Number of multi-start initials. 
%       -> numInits.rrr = Array of desired ranks to be used for intials 
%                         that are obtained from RRR/RRL. 
%                         For ordinary matrix models, the default is [r, q,
%                         m] with RRR (regardless of the loss), and for trace models, 
%                         RRL will be used, with the default rank at q.
%                         No initials obtained from RRR if it is empty.
%       -> numInits.rand = Number of random initials. (Default: 0).
%   estimateIntercept: (Default: 1)
%       0: No intercept estimation (return zeros).
%       1: The intercept alpha will be simultaneously estimated.      
%
% Output arguments:
%   B_inits: A 3D array (of size n x m x numInits) of the coefficient matrix 
%            estimates.
%   alpha_inits: A matrix (of size intrcptNo x numInits) of the intercept estimates.
%   S_inits: A 3D array (of size p x r x numInits) of the loading matrices.
%   V_inits: A 3D array (of size m x r x numInits) of the rotation matrices.
%   F_inits: A 3D array (of size p x q x numInits) of the membership matrices.
%   mu_inits: A 3D array (of size q x r x numInits) of the centroid matrices.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if ~contains(lossType, 'tracemodel')
    [n, m] = size(Y); p = size(X, 2); 
    intrcptNo = m; % num of components in the intercept column
else
    if size(X, 3) > 1
        % The most general trace model: y_i (scalar) ~ < X(:, :, i), B > + alpha_i
        [n] = size(Y, 1); 
        p = size(X, 1); m = size(X, 2); % B is of size pxm
        intrcptNo = n; % num of components in the intercept column
    else
        % if size(X, 3) == 1, i.e., a two-dim design matrix, 
        % we assume a "sample-based trace model": y_i (scalar) ~ < X(i, :), B(i, :) >  + alpha_i
        [n] = size(Y, 1); p = size(X, 1); m = size(X, 2); % B is of size pxm (or nxp in this special case)
        intrcptNo = n;
    end
end

% 1. Set the default numbers of initials 
if ~contains(lossType, 'tracemodel')
    tempks = [r, q, m]; % default rank values to be used to get RRR estimates
else
    tempks = [q];
end
tempnumInits = struct('rrr', tempks, 'rand', 0);
if ~exist('numInits', 'var') || isempty(numInits) 
    numInits = tempnumInits;
else
    if ~isfield(numInits, 'rrr') % Only check if it does not exist. User can give empty field.
        numInits.rrr = tempnumInits.rrr;
    end
    if ~isfield(numInits, 'rand') || isempty(numInits.rand) 
        numInits.rand = tempnumInits.rand;
    end
end

% 2. Creating space to store initials.
totInits = numel(numInits.rrr) + numInits.rand;

B_inits = zeros(p, m, totInits);
alpha_inits = zeros(intrcptNo, totInits);
S_inits = zeros(p, r, totInits);
V_inits = zeros(m, r, totInits);
F_inits = zeros(p, q, totInits);
mu_inits = zeros(q, r, totInits);

% 3.-----initialization algorithm----------
initIdx = 0;


% a. RRR-related initial points % assuming a quadratic loss
ks = numInits.rrr; 
if numel(ks)
    if ~contains(lossType, 'tracemodel')
        % Ordinary matrix models:
        [Bs_rrr, ~, ~, ~, ~] = Func_RRR(X, Y, ks, estimateIntercept);
        %[Bs_rrr, ~, ~, ~] = Func_RRL(X, Y, ks, lossType, estimateIntercept, [], []);
        for indB = 1:numel(ks)
            initIdx = initIdx + 1;
            B_raw = Bs_rrr(:, :, indB);
            [B_init, alpha_init, S_init, V_init, F_init, mu_init] = ...
                Func_RoughClusterB_aux(X, Y, B_raw, q, r, lossType, estimateIntercept, []);
            B_inits(:, :, initIdx) = B_init;
            alpha_inits(:, initIdx) = alpha_init;
            S_inits(:, :, initIdx) = S_init;
            V_inits(:, :, initIdx) = V_init;
            F_inits(:, :, initIdx) = F_init;
            mu_inits(:, :, initIdx) = mu_init;
        end
    else % trace models
        if r >= min([q, p, m]) || q <= 2 % No further low rank or q is very small
            tempconvCtrl = []; % struct('maxiter', 1000, 'tol', 1e-04); %  struct('maxiter', 25, 'tol', 1e-02); %  
        else
            tempconvCtrl = struct('maxiter', 5, 'tol', 1e-02); %[];  struct('maxiter', 25, 'tol', 1e-02); %  
        end
        for indB = 1:numel(ks)
            initIdx = initIdx + 1;
            [B_raw, ~, ~, ~] = Func_RRL(X, Y, ks(indB), lossType, estimateIntercept, tempconvCtrl, []);
            
            [B_init, alpha_init, S_init, V_init, F_init, mu_init] = ...
                Func_RoughClusterB_aux(X, Y, B_raw, q, r, lossType, estimateIntercept, tempconvCtrl);
            B_inits(:, :, initIdx) = B_init;
            alpha_inits(:, initIdx) = alpha_init;
            S_inits(:, :, initIdx) = S_init;
            V_inits(:, :, initIdx) = V_init;
            F_inits(:, :, initIdx) = F_init;
            mu_inits(:, :, initIdx) = mu_init;
        end
    end
end
    
% b. Generate random initial points
prop = 0.75; 
if numel(numInits.rand)
    for i = 1:numInits.rand
        initIdx = initIdx + 1;
        if ~contains(lossType, 'tracemodel')
            % Ordinary matrix models:    
            % Generate a random B_raw matrix        
            subsetsize = floor(n * prop);
            subinds = randsample(n, subsetsize);
            X_sub = X(subinds, :);
            Y_sub = Y(subinds, :);
            [B_raw, ~, ~, ~] = Func_FullmodelMLE(X_sub, Y_sub, 1, 'gaussian', []); % ols
            [B_init, alpha_init, S_init, V_init, F_init, mu_init] =...
                Func_RoughClusterB_aux(X, Y, B_raw, q, r, lossType, estimateIntercept, []); 
        else % trace models
            % Random sample to create an initial F
            if r >= min([q, p, m]) || q <= 2 % No further low rank or q is very small
                tempconvCtrl = []; % struct('maxiter', 1000, 'tol', 1e-04); %  struct('maxiter', 25, 'tol', 1e-02); %  
            else
                tempconvCtrl = struct('maxiter', 5, 'tol', 1e-02); %[];  struct('maxiter', 25, 'tol', 1e-02); %  
            end            
            F_raw = datasample(eye(q), p);% of size p x q
%             if i == 1
%                 F_raw = [ones(n/2, 1), zeros(n/2, 1); zeros(n/2, 1), ones(n/2, 1)];
%             end
            [B_init, alpha_init, S_init, V_init, F_init, mu_init] =...
                Func_Calibration(X, Y, lossType, estimateIntercept, tempconvCtrl, r, F_raw);            
        end
        B_inits(:, :, initIdx) = B_init;
        alpha_inits(:, initIdx) = alpha_init;
        S_inits(:, :, initIdx) = S_init;
        V_inits(:, :, initIdx) = V_init;
        F_inits(:, :, initIdx) = F_init;
        mu_inits(:, :, initIdx) = mu_init;
    end
end
end
