function [TransformMat, EquiSpa_patt, LowRank_patt] = Func_PatternExtr(B, q, r, F) 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Pattern Extraction based on an equi-sparse and/or low-rank Estimator 
% New design matrix with only q features can be obtained by: Z=X*TransformMat. 
%
% Input arguments:
%   B: Equi-sparse and/or low rank coefficient matrix of size p x m.
%   q: Number of row-clusters in B.
%   r: Rank of B.
%   F: Equi-sparsity inducing membership matrix of size p x q. (Default: F
%      is extracted from B if F is not provided)
%   
% Output arguments:
%   TransformMat: The orthogonal structural pattern matrix of size p x r
%                 with q distinct rows.
%   EquiSpa_patt: The orthogonal equi-sparsity pattern matrix of size p x q
%   LowRank_patt: The orthogonal low rank pattern matrix of size q x r.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if ~exist('F', 'var') || isempty(F)
    [F, ~ ] = Func_kmeans_aux(B, q, B(1:q,:)); 
end

extrWay = 1; % 0; % 1 % %%1 gives more comprehensive info
    % 0: directly from B, 1: sequential way: first extract the row-wise 
    % pattern, and then the low-rank pattern, 2: both ways
if extrWay == 0 || extrWay == 2
    % B is assumed to have q different rows and rank r
    [tU, tS, tV] = svd(B, 'econ'); % should have rank r
    TransformMat0 = tU(:, 1:r); 
end
if extrWay == 1 || extrWay == 2
    EquiSpa_patt = normc(F); % normalize to have each col's l2 norm exactly 1
    % Now we get the first orthogonal matrix EquiSpa_patt, to chacterize the
    % row-wise equisparsity patterns
    [tU, tS, tV] = svd(EquiSpa_patt' * B, 'econ'); 
    LowRank_patt = tU(:, 1:r);
    TransformMat1 = EquiSpa_patt * LowRank_patt;
end


if extrWay == 2
    % Check if both methods are extracting the same patterns (which should be numerically equal).
    tmpsubspace1 = LowRank_patt * LowRank_patt'; 
    tmpsubspace0 = EquiSpa_patt' * TransformMat0;
    tmpsubspace0 = tmpsubspace0 * tmpsubspace0';
    if  max( svd(tmpsubspace0 - tmpsubspace1) )  > 1
        disp('Extracted patterns are different');
    else
        disp('Extracted patterns are same');
    end
    
end

if extrWay == 0 
    TransformMat = TransformMat0;
    EquiSpa_patt = [];
    LowRank_patt = [];
else
    TransformMat = TransformMat1;
end
end