function [Best, alphaest, Sest, Vest] = ...
    Func_RRL(X, Y, r, lossType, estimateIntercept, convCtrl, init, rho_user)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Reduced Rank Learning.
% This function fits a rank-constrained model given a general loss (grad Lip 1). 
% For matrix models, the implementation is by iterative RRR (and RRR for regression), 
% and for trace models, the implmentation is by iteratative PCA (linearization wrt B)
%
% Notes: 1. The input X must not include the intercept column.
%        2. The corresponding coefficient matrix B will not contain 
%           intercept estimate. We will provide a separate intercept
%           estimate alpha as a column vector.
%        3. That is, the loss of the model is l_0(1 * alpha' + X * B; Y). 
%        4. Low-rank structure is imposed only on the coefficient matrix.
%
% Input arguments:
%   X: the design matrix (no intercept column)
%   Y: the response matrix
%   r: the rank value of interest.
%   losstype: Could be 'gaussian', 'bernoulli', 'poisson', a robust loss, 
%             or any other loss of interest (not all implemented yet). 
%             Check Func_Loss_Grad for more details.
%   estimateIntercept: (Default: 1)
%       0: No intercept estimation (return zeros).
%       1: The intercept alpha will be simultaneously estimated.      
%   convCtrl: A structure array for convergence control
%       -> convCtrl.maxiter: maximum number of iterations allowed.
%                            (Default: 50.)
%       -> convCtrl.tol: error tolerance. (Default: 1e-04.)
%   init: Give initial values for the iterative algorithm.
%       -> init.B: Give the initial value of B. (Default: a zero matrix of
%                  size p x m [or a crude estimate from RRR for marix models].)
%       -> init.alpha: Give the initial value of alpha. (Default: zero vector 
%                      of size intrcptNo x 1 [or crude estimate from RRR].)
%   rho_user: User provided inverse stepsize parameter (if not existing or
%   empty, use the default)
% Output arguments:
%   Best: The optimal B estimate with rank less than r. 
%   alphaest: The intercept estimate vector of size intrcptNo x 1.
%   Sest: The loading matrix of size p x r.
%   Vest: the rotation matrix of size m x r satisfying V'V = I.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
debug = 0; %1; % 
% 1. Set default values to some input paremeters if not provided
if ~exist('rho_user', 'var') || isempty(rho_user)
    rho = [];
else
    rho = rho_user;
end
if ~exist('init', 'var') || isempty(init)
    init_manner = 'zeros'; % 'rrr' % 
    % 'zeros': use zeros for B and alpha. 'rrr': pretend the loss is quadratic
else
    init_manner = 'user';    
end
% Default values to control the convergence 
tempconvCtrl = struct('maxiter', 1000, 'tol', 1e-04);
if ~exist('convCtrl', 'var') || isempty(convCtrl)
    convCtrl = tempconvCtrl;
else
    if ~isfield(convCtrl, 'maxiter') || isempty(convCtrl.maxiter)
        convCtrl.maxiter = tempconvCtrl.maxiter;
    end
    if ~isfield(convCtrl, 'tol') || isempty(convCtrl.tol)
        convCtrl.tol = tempconvCtrl.tol;
    end
end
% Default value to control intercept estimation
if ~exist('estimateIntercept', 'var') || isempty(estimateIntercept)
    estimateIntercept = 1;
end

% 2. Preprocess and construct the initial values
if ~contains(lossType, 'tracemodel')
    [n, m] = size(Y); p = size(X, 2); 
    intrcptNo = m; % num of components in the intercept column
    temp_rUppBnd = min([rank(X, 1e-5), m]); % rankX = rank(X, 1e-5);
    if isempty(rho)
        rho = 1;  % stepsize is 1, since we assume the gradient of the loss is Lip 1, and in this case we will linearize the loss wrt the systematic component
    end
else
    if size(X, 3) > 1
        % The most general trace model: y_i (scalar) ~ < X(:, :, i), B > + alpha_i
        [n] = size(Y, 1); 
        p = size(X, 1); m = size(X, 2); % B is of size pxm
        intrcptNo = n; % num of components in the intercept column
    else
        % if size(X, 3) == 1, i.e., a two-dim design matrix, 
        % we assume a "sample-based trace model": y_i (scalar) ~ < X(i, :), B(i, :) >  + alpha_i
        [n] = size(Y, 1); p = size(X, 1); m = size(X, 2); % B is of size pxm (or nxp in this special case)
        intrcptNo = n;
    end  
    temp_rUppBnd = min([n, m, p]); 
    if isempty(rho)
        [rho] = Func_RhoRef_InvStepsize(X, lossType, estimateIntercept); % In this case, we will linearize the loss wrt B
        % TO BE UPDDATED: include the rho calc when the intercept exists?? (The
        % current way might still work when alpha exists)
    end
end
maxiter = convCtrl.maxiter;
tol = convCtrl.tol;
% If r is too large, truncate it. 
r = min([r, temp_rUppBnd]); 
% Construct the starting point
switch init_manner
    case 'user'
        if isfield(init, 'B') && ~isempty(init.B)
            B_init = init.B;
        else
            B_init = zeros(p, m);
        end
        if estimateIntercept == 1
            if isfield(init, 'alpha') && ~isempty(init.alpha)
                alpha_init = init.alpha;
            else 
                alpha_init = zeros(intrcptNo, 1);
            end
        else
            alpha_init = zeros(intrcptNo, 1);
        end
    case 'zeros'
        B_init = zeros(p, m);
        alpha_init = zeros(intrcptNo, 1);
    case 'rrr' 
        if contains(lossType, 'tracemodel')
            error('Wrong init_manner for trace models.')
        end
        if estimateIntercept == 1
            if r >= max([size(X, 2), size(Y, 2)])
                [B_init, alpha_init, ~, ~] = Func_FullmodelMLE(X, Y, estimateIntercept, 'gaussian', []);
            else
                [B_init, alpha_init, ~, ~, ~] = Func_RRR(X, Y, r, 1);
            end
        else
            if r >= max([size(X, 2), size(Y, 2)])
                [B_init, ~, ~, ~] = Func_FullmodelMLE(X, Y, estimateIntercept, 'gaussian', []);
            else
                [B_init, ~, ~, ~, ~] = Func_RRR(X, Y, r, 0);
            end
            alpha_init = zeros(intrcptNo, 1);
        end
    otherwise
        error('The method for initialization is not implemented yet.')
end

% 3. Run the iterative algorithm
k = 0;
alpha_cur = alpha_init;
B_cur = B_init;

if contains(lossType, 'tracemodel') && estimateIntercept ~= 0, warning('The reduced rank learning in the trace model with an intercept is NOT implemented yet'), end

while 1
    k = k + 1;
    if debug > 0, disp(['*************** RRL: loop index k = ', num2str(k), ' *******************']), end
    % Constructing pseudo Y.
    if ~contains(lossType, 'tracemodel')
        % In this case, we linearized the loss wrt the systematic component
        [Theta] = Func_Eval_Theta_SysComp(X, B_cur, alpha_cur, lossType); % For ordinary models, Theta = X * B_cur + alpha_cur';
        [~, G] = Func_Loss_Grad(Y, Theta, lossType, 1);
%         G = G + (1e-4) * Theta; % add a mild l2 shrinkage
        Y_tilde = Theta - G / rho; 

        % Fitting RRR (Y_tilde ~ X)  to get updated B and alpha
        if estimateIntercept == 1
            [B_new, alpha_new, S_new, V_new, ~] = Func_RRR(X, Y_tilde, r, 1);
        else
            [B_new, ~, S_new, V_new, ~] = Func_RRR(X, Y_tilde, r, 0);
            alpha_new = zeros(intrcptNo, 1);
        end
    else
        % For trace models, we linearized the loss wrt B
%         if estimateIntercept ~= 0, warning('The reduced rank learning in the trace model with an intercept is NOT implemented yet'), end
        [XtG, ~, GtOnes] = Func_GradB(X, Y, lossType, estimateIntercept, B_cur, alpha_cur);        
%         XtG = XtG + (1e-4) * B_cur; % add a mild l2 shrinkage
        % intercept update
        if estimateIntercept == 1
            alpha_new = alpha_cur - GtOnes / rho; % i.e., alpha_new' = alpha_cur' - grad_alpha / rho 
            % NOTE: The stepsize calculation did not consider alpha
        else
            alpha_new = zeros(intrcptNo, 1);
        end        
        % B update (PCA)
        Y_tilde = B_cur - XtG / rho;        
        [B_new, ~, S_new, V_new, ~] = Func_RRR(eye(size(B_cur, 1)), Y_tilde, r, 0); % identity design, no intercept (just PCA!)
    end
    if debug > 0, disp(['RRL: new function value - old: ', num2str( Func_Loss_Grad_Balpha(Y, X, B_new, alpha_new, lossType, 0) - Func_Loss_Grad_Balpha(Y, X, B_cur, alpha_cur, lossType, 0) )]), end
    if k == maxiter || max(max(abs(B_new-B_cur) ) ) < tol
        Best = B_new;
        alphaest = alpha_new;
        Sest = S_new;
        Vest = V_new;
        break;
    end
    alpha_cur = alpha_new;
    B_cur = B_new;
    
    
end
end

