function [rho_ref] = Func_RhoRef_InvStepsize(X, lossType, estimateIntercept, rhofactor)
% This function calculates a reference rho, or inverse stepsize, based on
% theoretical derivations (which might conservative).
% We assume the loss defined on the systematic component has been
% normalized such that its Lip parameter can take 1. 

% disp('Stepsize selection starting')
if ~exist('rhofactor', 'var') || isempty('rhofactor')
    rhofactor = 1;
end

if ~contains(lossType, 'tracemodel')
    % ordinary matrix model Y ~ X * B + 1 * alpha' if intercept exists
    if estimateIntercept == 1
        X_bar = [ones(size(X, 1), 1), X];
        rho_ref = 1 * norm(X_bar)^2 * rhofactor; % The loss has Lip-1 gradient
    else
        rho_ref = 1 * norm(X)^2 * rhofactor;
    end
else % In this case, we assume a trace model without intercept
    if estimateIntercept ~= 0, warning('The choice of rho in the trace model with an intercept is NOT implemented yet'), end
    if size(X, 3) > 1
        % The most general trace model: y_i (scalar) ~ < X(:, :, i), B >
        % By matrix calculus, Hessian_all  = \sum tVec * (tVec)^T
        Hessian_all = zeros(size(X, 1) * size(X, 2));
        for i = 1:size(X, 3)
            tVec = reshape(X(:, :, i)', [], 1);  % Alt: reshape(X(:, :, i), [], 1);
            % The first is for vec(dB^T): vec(dB^T) ^T * H * vec(dB^T); the latter is
            % for vec(dB): vec(dB) ^T * H * vec(dB)
            Hessian_all = Hessian_all + tVec * tVec';
        end
        rho_ref = 1 * norm(Hessian_all, 2) * rhofactor;
    else
        % if size(X, 3) == 1, i.e., a two-dim design matrix, we assume a sample-based model: y_i (scalar) ~ < X(i, :), B(i, :) >
        rho_ref = 1 * max(sum(X.^2, 2)) * rhofactor; % can be verified to be the same as that in the general case, but is caculated more efficiently here
    end
end
% disp('Stepsize selection done')
end

