function [X, Y, tstX, tstY, valX, valY] = Func_preprocess(X_raw, Y_raw, preprocess, tstX_raw, tstY_raw, valX_raw, valY_raw)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Preprocess Y and X in various ways. 
%
% Input arguments: 
%   X_raw: Raw design matrix.
%   Y_raw: Raw response matrix.
%   preprocess: A structure array to provide different options to preprocess
%               X and Y.
%       -> preprocess.Y: (Default: whitening)  
%           'none': Does not perform any preprocessing.
%           'centering': Performs column centering only.
%           'normalization': Performs column centering followed by scaling.
%           'whitening': Performs column centering followed by multivariate 
%                        decorrelation.
%       -> preprocess.X: (Default: normalization)  
%           'none': Does not perform any preprocessing.
%           'centering': Performs column centering only.
%           'normalization': Performs column centering followed by scaling.
%
% Output arguments: 
%   X: Preprocessed X. 
%   Y: Preprocessed Y. 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% 1. Set the default preprocessing options 
if ~exist('valX_raw', 'var') || isempty(valX_raw) || ~exist('valY_raw', 'var') || isempty(valY_raw)
    valData = false;
    valX = [];
    valY = [];
else
    valData = true;
end
if ~exist('tstX_raw', 'var') || isempty(tstX_raw) || ~exist('tstY_raw', 'var') || isempty(tstY_raw)
    tstData = false;
    tstX = [];
    tstY = [];
else
    tstData = true;
end

if ~exist('preprocess', 'var') || isempty(preprocess) 
    preprocess.Y = 'whitening';
    preprocess.X = 'normalization'; 
else
    if ~isfield(preprocess, 'Y') || isempty(preprocess.Y) 
        preprocess.Y = 'whitening';
    end
    if ~isfield(preprocess, 'X') || isempty(preprocess.X) 
        preprocess.X = 'normalization';
    end
end




% Paramters used to handle the possible singularity of Y'*Y
singTh = 1e-5; 
regvalCov = 1e-4;

% 2. ---------Preprocess Y----------------
switch lower(preprocess.Y)
    case 'none'
        Y = Y_raw;
        if tstData
            tstY = tstY_raw;
        end
        if valData
            valY = valY_raw;
        end  
    case 'centering'
        %-----------Y column centering--------------------
        colMeans_Y = mean(Y_raw);
        Y = Y_raw - colMeans_Y; 
        if tstData
            tstY = tstY_raw - colMeans_Y; 
        end
        if valData
            valY = valY_raw - colMeans_Y;
        end
    case 'normalization'
        %-----------Y column centering--------------------
        colMeans_Y = mean(Y_raw);
        Y = Y_raw - colMeans_Y;
        if tstData
            tstY = tstY_raw - colMeans_Y; 
        end
        if valData
            valY = valY_raw - colMeans_Y;
        end
        %----------Y normalization---------------------
        stdY = std(Y); % sqrt( sum( (Y - mean(Y)).^2 ) / (size(Y, 1)-1) )
        Y = bsxfun(@rdivide, Y, stdY); % make each column of Y have norm sqrt n
        if tstData
            tstY = bsxfun(@rdivide, tstY, stdY); 
        end
        if valData
            valY = bsxfun(@rdivide, valY, stdY); 
        end
    case 'whitening'
        %-----------Y column centering--------------------
        colMeans_Y = mean(Y_raw);
        Y = Y_raw - colMeans_Y;
        if tstData
            tstY = tstY_raw - colMeans_Y; 
        end
        if valData
            valY = valY_raw - colMeans_Y;
        end
        %------------Y whiten---------------------------
        n = size(Y_raw, 1); 
        Sigma_Y = Y'*Y/n;
        [U, D] = eig(Sigma_Y);
        diag_D = diag(D);
        idx = find(abs(diag_D) < singTh, 1);
        if isempty(idx)
            colTransf_Y = U * (diag( (diag_D).^(-1/2) ) ) * U';
            Y = Y * colTransf_Y;   % Sigma_Y is nonsingular.
            if tstData
                tstY = tstY * colTransf_Y;
            end
            if valData
                valY = valY * colTransf_Y;
            end
        else
            colTransf_Y = U * (diag( (diag_D + regvalCov).^(-1/2) ) ) * U';
            Y = Y * colTransf_Y;   % Sigma_Y is singular
            if tstData
                tstY = tstY * colTransf_Y;
            end
            if valData
                valY = valY * colTransf_Y;
            end
        end
    otherwise 
        error('Invalid input for preprocess.Y. Enter "centering", "normalization" or "whitening".');
end

% 3. ----------Preprocess X--------------- 
switch lower(preprocess.X)
    case 'none'
        X = X_raw;
        if tstData
            tstX = tstX_raw;
        end
        if valData
            valX = valX_raw;
        end 
    case 'scaling'
        stdX = sqrt(sum(X_raw.^2)/size(X_raw, 1));
        X = bsxfun(@rdivide, X_raw, stdX); % make each column of X have norm sqrt n        
        if tstData
            tstX = bsxfun(@rdivide, tstX_raw, stdX); 
        end
        if valData
            valX = bsxfun(@rdivide, valX_raw, stdX); 
        end  
    case 'centering'
        %-----------X column centering--------------------
        colMeans_X = mean(X_raw);
        X = X_raw - mean(X_raw);
        if tstData
            tstX = tstX_raw - colMeans_X; 
        end
        if valData
            valX = valX_raw - colMeans_X; 
        end
    case 'normalization'
        %-----------X column centering--------------------
        colMeans_X = mean(X_raw);
        X = X_raw - colMeans_X;
        if tstData
            tstX = tstX_raw - colMeans_X; 
        end
        if valData
            valX = valX_raw - colMeans_X;
        end
        %-----------X normalization--------------------
        stdX = std(X);
        X = bsxfun(@rdivide, X, stdX); % make each column of X have norm sqrt n
        if tstData
            tstX = bsxfun(@rdivide, tstX, stdX); 
        end
        if valData
            valX = bsxfun(@rdivide, valX, stdX); 
        end  
    otherwise
        error('Invalid input for preprocess.X. Enter "centering" or "normalization".');
end
        

end


